from portpy.photon.utils import *
from .ct import CT
from .structures import Structures
from .beam import Beams
from .influence_matrix import InfluenceMatrix
from .clinical_criteria import ClinicalCriteria
try:
    from portpy.photon.vmat_scp import Arcs
except ImportError:
    Arcs = None


class Plan:
    """
    A class representing a plan

    - **Attributes** ::

        :param beams: an object of class Beams that contains information about the beams_dict used in the treatment plan.
        :type beams: object
        :param structures: an object of class Structures that contains information about the structures present in the patient's CT scan.
        :type structures: object
        :param inf_matrix: object of class Influence matrix
        :type inf_matrix: object
        :param clinial_criteria: an object of class ClincialCriteria that contains information about the goals and constraints of the treatment plan
        :type clinical_criteria: object
        :param ct: dictionary containing ct metadata. It includes metadata about important ct parameters like resolution, ct in HU etc.
        :type ct: object
        :param arcs: an object of class Arcs containing information about number of arcs and arcs features
        :type arcs: object

    - **Methods** ::

        :get_prescription():
            Return prescription for the plan
        :get_num_of_fractions()
            Return number of fractions for the plan




    """

    def __init__(self, structs: Structures, beams: Beams, inf_matrix: InfluenceMatrix,
                 ct: CT = None, clinical_criteria: ClinicalCriteria = None, arcs: Arcs = None) -> None:
        """
        Creates an object of Plan class for the specified patient

        :param ct: object of class CT
        :param structs: object of class structures
        :param beams: object of class Beams
        :param inf_matrix: object of class Influence matrix
        :param arcs: object of class Arcs

        :Example:

        >>> my_plan = Plan(structs, beams, inf_matrix, ct, clinical_criteria)
        """

        self.beams: Beams = beams  # create beams attribute
        self.structures: Structures = structs  # create structures attribute
        self.ct: CT = ct  # create ct attribute containing ct information as dictionary
        self.inf_matrix: InfluenceMatrix = inf_matrix
        if clinical_criteria is not None:
            self.clinical_criteria: ClinicalCriteria = clinical_criteria
        if ct is not None:
            self.patient_id = ct.patient_id
        if arcs is not None:
            self.arcs: Arcs = arcs

    def save_plan(self, plan_name: str = None, path: str = None) -> None:
        """

        Save pickled file for plan object

        :param plan_name: create the name of the pickled file of plan object. If none, it will save with the name as 'my_plan'
        :param path: if path is set, plan object will be pickled and saved in path directory else it will save current project directory
        :return: save pickled object of class Plan

        :Example:
        >>> Plan.save_plan(plan_name='my_plan', path=r"path/to/save_plan")
        """
        save_plan(self, plan_name=plan_name, path=path)

    @staticmethod
    def load_plan(plan_name: str = None, path: str = None):
        """
        Load pickle file of the plan object.

        :param plan_name: plan_name of the object of class Plan. It None, it will try to look for plan name called 'my_plan'
        :param path: if path is set, plan object will be load from path directory else current project directory
        :return: load pickled object of class Plan

        :Example:
        >>> Plan.load_plan(plan_name='my_plan', path=r"path/for/loading_plan")
        """

        return load_plan(plan_name=plan_name, path=path)

    @staticmethod
    def load_optimal_sol(sol_name: str, path: str = None) -> dict:
        """

        Load optimal solution dictionary got from optimization

        :param sol_name: name of the optimal solution to be loaded.
        :param path: if path is set, plan object will be load from path directory else current directory
        :return: load solution

        :Example:
        >>> sol = Plan.load_optimal_sol(sol_name='sol', path=r'path/for/loading_sol')
        """
        return load_optimal_sol(sol_name=sol_name, path=path)

    @staticmethod
    def save_optimal_sol(sol: dict, sol_name: str, path: str = None) -> None:
        """
        Save the optimal solution dictionary from optimization

        :param sol: optimal solution dictionary
        :param sol_name: name of the optimal solution saved
        :param path: if path is set, plan object will be load from path directory else current directory
        :return: save pickled file of optimal solution dictionary

        :Example:
        >>> Plan.save_optimal_sol(sol=sol, sol_name='sol', path=r'path/to/save_solution')
        """
        save_optimal_sol(sol=sol, sol_name=sol_name, path=path)

    def get_prescription(self) -> float:
        """

        :return: prescription in Gy
        """
        pres = self.clinical_criteria.clinical_criteria_dict['pres_per_fraction_gy'] * \
               self.clinical_criteria.clinical_criteria_dict[
                   'num_of_fractions']
        return pres

    def get_num_of_fractions(self) -> float:
        """

        :return: number of fractions to be delivered
        """
        return self.clinical_criteria.clinical_criteria_dict['num_of_fractions']

    def get_disease_site(self) -> float:
        """

        :return: number of fractions to be delivered
        """
        return self.clinical_criteria.clinical_criteria_dict['disease_site']

    def save_nrrd(self, sol: dict, data_dir: str = None):
        """
        save nrrd in the path directory else save in patient data directory
        :param sol: optimal solution dict
        :param data_dir: save nrrd images of ct, dose_1d and struct_name set in path directory
        :return: save nrrd images in path
        """
        save_nrrd(self, sol=sol, data_dir=data_dir)


from __future__ import annotations
from typing import List
import pandas as pd
from .data_explorer import DataExplorer
import json
from typing import TYPE_CHECKING, Union
if TYPE_CHECKING:
    from portpy.photon.plan import Plan
from copy import deepcopy
import numpy as np


class ClinicalCriteria:
    """
    A class representing clinical criteria.

    - **Attributes** ::

        :param clinical_criteria_dict: dictionary containing metadata about clinical criteria

        :Example
                dict = {"disease_site": "Lung",
                "protocol_name": "Lung_Default_2Gy_30Fr",
                "pres_per_fraction_gy": 2,
                "num_of_fractions": 30,
                "criteria": [
                    {
                      "type": "max_dose",
                      "parameters": {
                        "struct": "GTV"
                      }}

    - **Methods** ::
        :add_criterion(criterion:str, parameters:dict, constraints:dict)
        :modify_criterion(criterion:str, parameters:dict, constraints:dict)

    """

    def __init__(self, data: DataExplorer = None, protocol_name: str = None, protocol_type: str = 'Default', file_name: str = None):
        """

        :param clinical_criteria: dictionary containing information about clinical criteria

        """
        if file_name is None:
            clinical_criteria_dict = data.load_config_clinical_criteria(protocol_name, protocol_type=protocol_type)
        else:
            f = open(file_name)
            clinical_criteria_dict = json.load(f)
            f.close()
        self.clinical_criteria_dict = clinical_criteria_dict
        self.dvh_table = pd.DataFrame()

    def get_prescription(self) -> float:
        """

        :return: prescription in Gy
        """
        pres = self.clinical_criteria_dict['pres_per_fraction_gy'] * \
               self.clinical_criteria_dict[
                   'num_of_fractions']
        return pres

    def get_num_of_fractions(self) -> float:
        """

        :return: number of fractions to be delivered
        """
        return self.clinical_criteria_dict['num_of_fractions']

    def add_criterion(self, type: str, parameters: dict, constraints: dict) -> None:
        """
        Add criterion to the clinical criteria dictionary

        :param type: criterion name. e.g. max_dose
        :param parameters: parameters dictionary e.g. parameters = {'struct':'PTV'}
        :param constraints: constraints dictionary e.g. constraints = {'limit_dose_gy':66, 'goal_dose_gy':60}
        :return: add the criteria to clinical criteria dictionary

        """

        self.clinical_criteria_dict['criteria'].append({'type': type})
        self.clinical_criteria_dict['criteria'][-1]['parameters'] = parameters
        self.clinical_criteria_dict['criteria'][-1]['constraints'] = constraints

    @staticmethod
    def create_criterion(type: str, parameters: dict, constraints: dict):
        """
        Create criterion and return list

        :param type: criterion name. e.g. max_dose
        :param parameters: parameters dictionary e.g. parameters = {'struct':'PTV'}
        :param constraints: constraints dictionary e.g. constraints = {'limit_dose_gy':66, 'goal_dose_gy':60}
        :return: add the criteria to clinical criteria dictionary

        """

        criterion = [{'type': type, 'parameters': parameters, 'constraints': constraints}]
        return criterion

    def get_criteria(self, type: str = None) -> List[dict]:
        """
        Returns all the clinical criteria
        :return:
        """
        all_criteria = []
        if type is None:
            all_criteria = self.clinical_criteria_dict['criteria']
        elif type == 'max_dose':
            criteria = self.clinical_criteria_dict['criteria']
            ind = [i for i in range(len(criteria)) if criteria[i]['type'] == type]
            if len(ind) > 0:
                all_criteria = [criteria[i] for i in ind]
        elif type == 'mean_dose':
            criteria = self.clinical_criteria_dict['criteria']
            ind = [i for i in range(len(criteria)) if criteria[i]['type'] == type]
            if len(ind) > 0:
                all_criteria = [criteria[i] for i in ind]

        if isinstance(all_criteria, dict):
            all_criteria = [all_criteria]
        return all_criteria

    def check_criterion_exists(self, criterion, return_ind:bool = False):
        criterion_exist = False
        criterion_ind = None
        for ind, crit in enumerate(self.clinical_criteria_dict['criteria']):
            if (crit['type'] == criterion['type']) and crit['parameters'] == criterion['parameters']:
                for constraint in crit['constraints']:
                    if constraint == criterion['constraints']:
                        criterion_exist = True
                        criterion_ind = ind
        if return_ind:
            return criterion_exist,criterion_ind
        else:
            return criterion_exist

    def modify_criterion(self, criterion):
        """
        Modify the criterion the clinical criteria


        """
        criterion_found = False
        for ind, crit in enumerate(self.clinical_criteria_dict['criteria']):
            if (crit['type'] == criterion['type']) and crit['parameters'] == criterion['parameters']:
                for constraint in crit['constraints']:
                    if constraint == criterion['constraints']:
                        self.clinical_criteria_dict['criteria'][ind]['constraints'][constraint] = criterion['constraints']
                        criterion_found = True
        if not criterion_found:
            raise Warning('No criteria  for {}'.format(criterion))


    def get_num(self, string: Union[str, float]):
        if "prescription_gy" in str(string):
            prescription_gy = self.get_prescription()
            return eval(string)
        elif isinstance(string, float) or isinstance(string, int):
            return string
        else:
            raise Exception('Invalid constraint')

    def matching_keys(self, dictionary, search_string):
        get_key = None
        for key, val in dictionary.items():
            if search_string in key:
                get_key = key
        if get_key is not None:
            return get_key
        else:
            return ''

    def dose_to_gy(self, key, value):
        if "prescription_gy" in str(value):
            prescription_gy = self.get_prescription()
            return eval(value)
        elif 'gy' in key:
            return value
        elif 'perc' in key:
            return value*self.get_prescription()/100

    @staticmethod
    def convert_dvh_to_dose_gy_vol_perc(my_plan, old_criteria):
        """
        Get dose volume in Gy and Percent format
        """
        criteria = deepcopy(old_criteria)
        struct_name = criteria['parameters']['structure_name']
        if criteria['type'] == 'dose_volume_D':
            constraint_keys = criteria['constraints'].keys()
            for key in constraint_keys:
                if 'perc' in key:
                    value = criteria['constraints'][key]
                    new_key = key.replace('perc', 'gy')
                    criteria['constraints'][new_key] = criteria['constraints'].pop(key)
                    criteria['constraints'][new_key] = value / 100 * my_plan.get_prescription()
            param_keys = criteria['parameters'].keys()
            for key in param_keys:
                if 'volume_cc' in key:
                    value = criteria['parameters'][key]
                    new_key = key.replace('cc', 'perc')
                    criteria['parameters'][new_key] = criteria['parameters'].pop(key)
                    criteria['parameters'][new_key] = value / my_plan.structures.get_volume_cc(
                        struct_name.upper()) * 100
        if criteria['type'] == 'dose_volume_V':
            constraint_keys = criteria['constraints'].keys()
            for key in constraint_keys:
                if 'volume_cc' in key:
                    value = criteria['constraints'][key]
                    new_key = key.replace('cc', 'perc')
                    criteria['constraints'][new_key] = criteria['constraints'].pop(key)
                    criteria['constraints'][new_key] = value / my_plan.structures.get_volume_cc(
                        struct_name.upper()) * 100
            param_keys = criteria['parameters'].keys()
            for key in param_keys:
                if 'dose_perc' in key:
                    value = criteria['parameters'][key]
                    new_key = key.replace('perc', 'gy')
                    criteria['parameters'][new_key] = criteria['parameters'].pop(key)
                    criteria['parameters'][new_key] = value / 100 * my_plan.get_prescription()
        return criteria


    def get_dvh_table(self, my_plan: Plan, constraint_list: list = None, opt_params: Union[list, dict] = None):
        if constraint_list is None:
            constraint_list = deepcopy(self.clinical_criteria_dict['criteria'])
        if opt_params is not None:
            # add/modify constraints definition if present in opt params
            for opt_constraint in opt_params['constraints']:
                # add constraint
                param = opt_constraint['parameters']
                if param['structure_name'] in my_plan.structures.get_structures():
                    criterion_exist, criterion_ind = self.check_criterion_exists(opt_constraint,
                                                                                              return_ind=True)
                    if criterion_exist:
                        constraint_list[criterion_ind] = opt_constraint
                    else:
                        constraint_list += [opt_constraint]

        dvh_updated_list = []
        for i, constraint in enumerate(constraint_list):
            if constraint['parameters']['structure_name'] in my_plan.structures.get_structures():
                if len(my_plan.inf_matrix.get_opt_voxels_idx(constraint['parameters']['structure_name'])) == 0:
                    continue
                updated_constraint = self.convert_dvh_to_dose_gy_vol_perc(my_plan, constraint)
                dvh_updated_list.append(updated_constraint)
        import pandas as pd
        df = pd.DataFrame()
        count = 0
        for i in range(len(dvh_updated_list)):
            if 'dose_volume_V' in dvh_updated_list[i]['type']:
                limit_key = self.matching_keys(dvh_updated_list[i]['constraints'], 'limit')
                dose_key = self.matching_keys(dvh_updated_list[i]['parameters'], 'dose_')
                if limit_key in dvh_updated_list[i]['constraints']:
                    df.at[count, 'structure_name'] = dvh_updated_list[i]['parameters']['structure_name']
                    df.at[count, 'dose_gy'] = self.dose_to_gy(dose_key, dvh_updated_list[i]['parameters'][dose_key])
                    df.at[count, 'volume_perc'] = dvh_updated_list[i]['constraints'][limit_key]
                    df.at[count, 'dvh_type'] = 'constraint'
                    count = count + 1
                goal_key = self.matching_keys(dvh_updated_list[i]['constraints'], 'goal')
                if goal_key in dvh_updated_list[i]['constraints']:
                    df.at[count, 'structure_name'] = dvh_updated_list[i]['parameters']['structure_name']
                    df.at[count, 'dose_gy'] = self.dose_to_gy(dose_key, dvh_updated_list[i]['parameters'][dose_key])
                    df.at[count, 'volume_perc'] = dvh_updated_list[i]['constraints'][goal_key]
                    df.at[count, 'dvh_type'] = 'goal'
                    count = count + 1
            if 'dose_volume_D' in dvh_updated_list[i]['type']:
                limit_key = self.matching_keys(dvh_updated_list[i]['constraints'], 'limit')
                if limit_key in dvh_updated_list[i]['constraints']:
                    df.at[count, 'structure_name'] = dvh_updated_list[i]['parameters']['structure_name']
                    df.at[count, 'volume_perc'] = dvh_updated_list[i]['parameters']['volume_perc']
                    df.at[count, 'dose_gy'] = self.dose_to_gy(limit_key, dvh_updated_list[i]['constraints'][limit_key])
                    df.at[count, 'dvh_type'] = 'constraint'
                    count = count + 1
                goal_key = self.matching_keys(dvh_updated_list[i]['constraints'], 'goal')
                if goal_key in dvh_updated_list[i]['constraints']:
                    df.at[count, 'structure_name'] = dvh_updated_list[i]['parameters']['structure_name']
                    df.at[count, 'volume_perc'] = dvh_updated_list[i]['parameters']['volume_perc']
                    df.at[count, 'dose_gy'] = self.dose_to_gy(goal_key, dvh_updated_list[i]['constraints'][goal_key])
                    df.at[count, 'dvh_type'] = 'goal'
                    count = count + 1
        self.dvh_table = df
        self.get_max_tol(constraints_list=constraint_list)
        return self.dvh_table


    def get_low_dose_vox_ind(self, my_plan: Plan, dose: np.ndarray):
        dvh_table = self.dvh_table
        inf_matrix = my_plan.inf_matrix
        for ind in dvh_table.index:
            structure_name, dose_gy, vol_perc = dvh_table['structure_name'][ind], dvh_table['dose_gy'][ind], \
            dvh_table['volume_perc'][ind]
            dvh_type = dvh_table['dvh_type'][ind]
            vol_perc = vol_perc / inf_matrix.get_fraction_of_vol_in_calc_box(structure_name)
            struct_vox = inf_matrix.get_opt_voxels_idx(structure_name)
            n_struct_vox = len(struct_vox)
            sort_ind = np.argsort(dose[struct_vox])
            voxel_sort = struct_vox[sort_ind]
            weights = inf_matrix.get_opt_voxels_volume_cc(structure_name)
            weights_sort = weights[sort_ind]
            weight_all_sum = np.sum(weights_sort)
            w_sum = 0
            if dvh_type == 'constraint':
                for w_ind in range(n_struct_vox):
                    w_sum = w_sum + weights_sort[w_ind]
                    w_ratio = w_sum / weight_all_sum
                    if w_ratio * 100 >= (100 - vol_perc):
                        break
                low_dose_voxels = voxel_sort[:w_ind+1]
                if ind == 0:
                    dvh_table.at[ind, 'low_dose_voxels'] = object  # fix issue with adding array to dataframe
                dvh_table.at[ind, 'low_dose_voxels'] = low_dose_voxels

        return self.dvh_table

    def get_max_tol(self, constraints_list: list = None):
        if constraints_list is None:
            constraints_list = self.clinical_criteria_dict['criteria']
        dvh_table = self.dvh_table
        for ind in dvh_table.index:
            structure_name, dose_gy = dvh_table['structure_name'][ind], dvh_table['dose_gy'][ind]
            max_tol = 100
            for criterion in constraints_list:
                if criterion['type'] == 'max_dose':
                    if criterion['parameters']['structure_name'] == structure_name:
                        key = self.matching_keys(criterion['constraints'], 'limit')
                        max_tol = self.dose_to_gy(key, criterion['constraints'][key])
            dvh_table.at[ind, 'max_tol'] = max_tol

        return self.dvh_table


from scipy import interpolate
import numpy as np
import pandas as pd
import webbrowser
from typing import List, Union
import os

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from .plan import Plan
from .clinical_criteria import ClinicalCriteria
from tabulate import tabulate


class Evaluation:
    """
    Class for evaluating different metrics for the plan

    - **Methods** ::
        :display_clinical_criteria(my_plan, sol)
            display plan values corresponding to given clinical criteria
        :get_dose(sol, struct, volume_per)
            get dose at the given volume in percentage
        :get_volume(sol: struct dose_value_gy)
            Get volume at dose_1d value in Gy


    """

    @staticmethod
    def display_clinical_criteria(my_plan: Plan, sol: Union[dict, List[dict]] = None, dose_1d: Union[np.ndarray, List[np.ndarray]]=None, html_file_name='temp.html',
                                  sol_names: List[str] = None, clinical_criteria: ClinicalCriteria = None,
                                  return_df: bool = False, in_browser: bool = False, path: str = None, open_browser: bool = True):
        """
        Visualization the plan metrics for clinical criteria in browser.
        It evaluate the plan by comparing the metrics against required criteria.

        If plan value is green color. It meets all the Limits and Goals
        If plan value is yellow color. It meets limits but not goals
        If plan value is red color. It violates both limit and goals

        :param my_plan: object of class Plan
        :param sol: optimal solution dictionary
        :param dose_1d: vectorized dose 1d array
        :param html_file_name:  name of the html file to be launched in browser
        :param sol_names: Default to Plan Value. column names for the plan evaluation
        :param clinical_criteria: clinical criteria to be evaluated
        :param return_df: return df instead of visualization
        :param in_browser: display table in browser
        :param path: path for saving the html file which opens up in browser
        :param open_browser: if true, html will be launched in browser
        :return: plan metrics in browser
        """
        import re
        # convert clinical criteria in dataframe
        if clinical_criteria is None:
            clinical_criteria = my_plan.clinical_criteria
        # df = pd.DataFrame.from_dict(clinical_criteria.clinical_criteria_dict['criteria'])
        df = pd.json_normalize(clinical_criteria.clinical_criteria_dict['criteria'])
        dose_volume_V_ind = df.index[df['type'] == 'dose_volume_V'].tolist()
        if dose_volume_V_ind:
            volumn_cols = [col for col in df.columns if 'volume' in col]
            if volumn_cols:
                perc_col = [col_name for col_name in volumn_cols if 'perc' in col_name]
                cc_col = [col_name for col_name in volumn_cols if 'cc' in col_name]
                for col in perc_col:
                    if 'limit' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'Limit', col, '%')
                    if 'goal' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'Goal', col, '%')
                for col in cc_col:
                    if 'limit' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'Limit', col, 'cc')
                    if 'goal' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'Goal', col, 'cc')
            dose_cols = [col for col in df.columns if 'parameters.dose' in col]
            if dose_cols:
                for col in dose_cols:
                    if 'perc' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'type', col, '%', 'dose_volume_V')
                    if 'gy' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'type', col, 'Gy', 'dose_volume_V')

        dose_cols = [col for col in df.columns if 'dose' in col]
        if dose_cols:
            perc_col = [col_name for col_name in dose_cols if 'perc' in col_name]
            gy_col = [col_name for col_name in dose_cols if 'gy' in col_name]
            for col in perc_col:
                if 'limit' in col:
                    df = Evaluation.add_dvh_to_frame(my_plan, df, 'Limit', col, '%')
                if 'goal' in col:
                    df = Evaluation.add_dvh_to_frame(my_plan, df, 'Goal', col, '%')
            for col in gy_col:
                if 'limit' in col:
                    df = Evaluation.add_dvh_to_frame(my_plan, df, 'Limit', col, 'Gy')
                if 'goal' in col:
                    df = Evaluation.add_dvh_to_frame(my_plan, df, 'Goal', col, 'Gy')

        dose_volume_D_ind = df.index[df['type'] == 'dose_volume_D'].tolist()
        if dose_volume_D_ind:
            vol_cols = [col for col in df.columns if 'parameters.volume' in col]
            if vol_cols:
                for col in vol_cols:
                    if 'perc' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'type', col, '%', 'dose_volume_D')
                    if 'cc' in col:
                        df = Evaluation.add_dvh_to_frame(my_plan, df, 'type', col, 'cc', 'dose_volume_D')

        # refine df
        df = df.rename(columns={'parameters.structure_name': 'structure_name', 'type': 'constraint'})
        # df = df.drop(
        #     ['parameters.dose_gy', 'constraints.limit_dose_gy', 'constraints.limit_volume_perc',
        #      'constraints.goal_dose_gy', 'constraints.goal_volume_perc','parameters.structure_def'], axis=1, errors='ignore')
        if 'Goal' not in df:
            df['Goal'] = ''
        df = df[['constraint', 'structure_name', 'Limit', 'Goal']]

        dose_1d_list = []
        dummy_sol = {}
        if isinstance(sol, dict):
            sol = [sol]
        if dose_1d is None:
            for p, s in enumerate(sol):
                dose_1d_list.append(s['inf_matrix'].A @ (s['optimal_intensity'] * my_plan.get_num_of_fractions()))
        else:
            if isinstance(dose_1d, np.ndarray):
                dose_1d_list = [dose_1d]
            else:
                dose_1d_list = dose_1d
        if sol_names is None:
            if len(dose_1d_list) > 1:
                sol_names = ['Plan Value ' + str(i) for i in range(len(dose_1d_list))]
            else:
                sol_names = ['Plan Value']
        for p, dose_1d in enumerate(dose_1d_list):
            dummy_sol['inf_matrix'] = my_plan.inf_matrix
            dummy_sol['dose_1d'] = dose_1d
            for ind in range(len(df)):  # Loop through the clinical criteria
                if df.constraint[ind] == 'max_dose':
                    struct = df.structure_name[ind]
                    if struct in my_plan.structures.get_structures():

                        max_dose = Evaluation.get_max_dose(dummy_sol, dose_1d=dose_1d, struct=struct)  # get max dose_1d
                        if 'Gy' in str(df.Limit[ind]) or 'Gy' in str(df.Goal[ind]):
                            df.at[ind, sol_names[p]] = np.round(max_dose,2)
                        elif '%' in str(df.Limit[ind]) or '%' in str(df.Goal[ind]):
                            df.at[ind, sol_names[p]] = np.round(max_dose / my_plan.get_prescription() * 100, 2)
                elif df.constraint[ind] == 'mean_dose':
                    struct = df.structure_name[ind]
                    if struct in my_plan.structures.get_structures():
                        mean_dose = Evaluation.get_mean_dose(dummy_sol, dose_1d=dose_1d, struct=struct)
                        if 'Gy' in str(df.Limit[ind]) or 'Gy' in str(df.Goal[ind]):
                            df.at[ind, sol_names[p]] = np.round(mean_dose, 2)
                        elif '%' in str(df.Limit[ind]) or '%' in str(df.Goal[ind]):
                            df.at[ind, sol_names[p]] = np.round(mean_dose / my_plan.get_prescription() * 100, 2)
                elif "V(" in df.constraint[ind]:
                    struct = df.structure_name[ind]
                    if struct in my_plan.structures.get_structures():
                        dose = re.findall(r"[-+]?(?:\d*\.*\d+)", df.constraint[ind])[0]
                        # convert dose to Gy
                        if '%' in df.constraint[ind]:
                            dose = float(dose)
                            dose = dose * my_plan.get_prescription() / 100
                        elif 'Gy' in df.constraint[ind]:
                            dose = float(dose)
                        # get volume in perc
                        volume = Evaluation.get_volume(dummy_sol, dose_1d=dose_1d, struct=struct, dose_value_gy=dose)
                        if '%' in str(df.Limit[ind]) or '%' in str(df.Goal[ind]): # we are writing str since nan values throws error
                            df.at[ind, sol_names[p]] = np.round(volume, 2)
                        elif 'cc' in str(df.Limit[ind]) or 'cc' in str(df.Goal[ind]):
                            vol_cc = my_plan.structures.get_volume_cc(structure_name=struct) * volume / 100
                            df.at[ind, sol_names[p]] = np.round(vol_cc, 2)
                elif "D(" in df.constraint[ind]:
                    struct = df.structure_name[ind]
                    if struct in my_plan.structures.get_structures():
                        volume = re.findall(r"[-+]?(?:\d*\.*\d+)", df.constraint[ind])[0]
                        # convert volume to perc
                        if '%' in df.constraint[ind]:
                            volume = float(volume)
                        elif 'cc' in df.constraint[ind]:
                            volume = float(volume)
                            volume = volume / my_plan.structures.get_volume_cc(structure_name=struct) * 100

                        # get dose
                        dose = Evaluation.get_dose(dummy_sol, dose_1d=dose_1d, struct=struct, volume_per=volume)
                        if '%' in str(df.Limit[ind]) or '%' in str(df.Goal[ind]): # we are writing str since nan values throws error
                            df.at[ind, sol_names[p]] = np.round(dose/my_plan.get_prescription()*100, 2)
                        elif 'Gy' in str(df.Limit[ind]) or 'Gy' in str(df.Goal[ind]):
                            df.at[ind, sol_names[p]] = np.round(dose, 2)
        df.round(2)
        df = df[df[sol_names].notna().all(axis=1)]  # remove rows for which plan value is Nan
        df = df.fillna('')
        # df.dropna(axis=0, inplace=True)  # remove structures which are not present
        # df.reset_index(drop=True, inplace=True)  # reset the index

        def color_plan_value(row):

            highlight_red = 'background-color: red;'   # red
            highlight_green = 'background-color: #90ee90'  # green
            highlight_orange = 'background-color: #ffb38a'  # orange
            default = ''

            row_color = len(row) * [default]  # default color for all rows initially
            # must return one string per cell in this row
            for i in range(len(row) - 2):
                if 'Limit' in row:
                    if not row['Limit'] == '':
                        limit = float(re.findall(r"[-+]?(?:\d*\.*\d+)", row['Limit'])[0])
                        if row[i] > limit + 0.0001:  # added epsilon to avoid minor differences
                            row_color[i] = highlight_red  # make plan value in red
                        else:
                            row_color[i] = highlight_green  # make plan value in red
                if 'Goal' in row:
                    if not row['Goal'] == '':
                        goal = float(re.findall(r"[-+]?(?:\d*\.*\d+)", row['Goal'])[0])
                        if row[i] > goal + 0.0001:
                            row_color[i] = highlight_orange  # make plan value in red
                        else:
                            row_color[i] = highlight_green  # make plan value in red
            return row_color

        sol_names.append('Limit')
        sol_names.append('Goal')
        df.style.set_properties(**{'text-align': 'right'})
        styled_df = df.style.apply(color_plan_value, subset=sol_names, axis=1)  # apply
        styled_df.set_properties(**{'text-align': 'center'}).format(precision=2)

        # color to dataframe using df.style method
        if return_df:
            return styled_df
        if in_browser:
            html = styled_df.render()  # render to html
            html_string = '''
                                                    <html>
                                                      <head><title>Portpy Clinical Criteria Evaluation</title></head>
                                                      <style> 
                                                        table, th, td {{font-size:10pt; border:1px solid black; border-collapse:collapse; text-align:left;}}
                                                        th, td {{padding: 5px;}}
                                                      </style>
                                                      <body>
                                                      <h1> Clinical Criteria</h1>
                                                      <h4 style="color: #90ee90">Meets limit and goal</h4>
                                                      <h4 style="color: #ffb38a">Meets limit but not goal</h4>
                                                      <h4 style="color: red">Violate both limit and goal</h4>
                                                        {table}
                                                      </body>
                                                    </html>.
                                                    '''
            if path is None:
                path = os.getcwd()
            html_file_path = os.path.join(path, html_file_name)
            with open(html_file_path, 'w') as f:
                f.write(html_string.format(table=html))
            if open_browser:
                webbrowser.open('file://' + os.path.realpath(html_file_path))

        else:
            if Evaluation.is_notebook():
                from IPython.display import display
                with pd.option_context('display.max_rows', None,
                                       'display.max_columns', None,
                                       'display.precision', 3,
                                       ):

                    display(styled_df)
            else:
                print(tabulate(df, headers='keys', tablefmt='psql'))  # print in console using tabulate

    @staticmethod
    def get_dose(sol: dict, struct: str, volume_per: float, dose_1d: np.ndarray = None,
                 weight_flag: bool = True) -> float:
        """
        Get dose_1d at volume percentage

        :param sol: solution dictionary
        :param dose_1d: dose_1d in 1d
        :param struct: struct_name name for which to get the dose_1d
        :param volume_per: query the dose_1d at percentage volume
        :param weight_flag: for non uniform voxels weight flag always True
        :return: dose_1d at volume_percentage

        :Example:

        >>> Evaluation.get_dose(sol=sol, struct='PTV', volume_per=90)

        """
        x, y = Evaluation.get_dvh(sol, dose_1d=dose_1d, struct=struct, weight_flag=weight_flag)
        if np.array_equal(x, np.array([0])) and np.array_equal(y, np.array([0])):
            return 0
        f = interpolate.interp1d(100 * y, x)
        if volume_per > 100.1:
            print('Warning: Volume Percentage: {} for structure {} is invalid'.format(volume_per, struct))
            return 0
        return f(volume_per)

    @staticmethod
    def get_volume(sol: dict, struct: str, dose_value_gy: float, dose_1d: np.ndarray = None,
                   weight_flag: bool = True) -> float:
        """
        Get volume at dose_1d value in Gy

        :param sol: solution dictionary
        :param dose_1d: dose_1d in 1d
        :param struct: struct_name name for which to get the dose_1d
        :param dose_value_gy: query the volume at dose_value
        :param weight_flag: for non uniform voxels weight flag always True
        :return: dose_1d at volume_percentage

        :Example:

        >>> Evaluation.get_volume(sol=sol, struct='PTV', dose_value_gy=60)

        """
        x, y = Evaluation.get_dvh(sol, dose_1d=dose_1d, struct=struct, weight_flag=weight_flag)
        if np.array_equal(x, np.array([0])) and np.array_equal(y, np.array([0])):
            return 0
        x1, indices = np.unique(x, return_index=True)
        y1 = y[indices]
        f = interpolate.interp1d(x1, 100 * y1)
        if dose_value_gy > max(x1):
            print('Warning: dose_1d value {} is greater than max dose_1d for {}'.format(dose_value_gy, struct))
            return 0
        else:
            return f(dose_value_gy)

    @staticmethod
    def get_dvh(sol: dict, struct: str, dose_1d: np.ndarray = None, weight_flag: bool = True):
        """
        Get dvh for the struct_name

        :param sol: optimal solution dictionary
        :param dose_1d: dose_1d which is not in solution dictionary
        :param struct: struct_name name
        :param weight_flag: for non uniform voxels weight flag always True
        :return: x, y --> dvh for the struct_name

        :Example:

        >>> Evaluation.get_dvh(sol=sol, struct='PTV')

        """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if len(vox) == 0:
            return np.array([0]), np.array([0])  # bug fix. if single 0 it can throw error while doing interpolation
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        org_sort_dose = np.sort(dose_1d[vox])
        sort_ind = np.argsort(dose_1d[vox])
        org_sort_dose = np.append(org_sort_dose, org_sort_dose[-1] + 0.01)
        x = org_sort_dose
        if weight_flag:
            # org_points_sort_spacing = my_plan._structures.opt_voxels_dict['dose_voxel_resolution_XYZ_mm']
            # org_points_sort_volume = org_points_sort_spacing[:, 0] * org_points_sort_spacing[:,
            #                                                          1] * org_points_sort_spacing[:, 2]
            # sum_weight = np.sum(org_points_sort_volume)
            org_weights = inf_matrix.get_opt_voxels_volume_cc(struct)
            org_sort_weights = org_weights[sort_ind]
            sum_weight = np.sum(org_sort_weights)
            frac_vol = inf_matrix.get_fraction_of_vol_in_calc_box(struct)
            if frac_vol is None:
                y = [1]
            else:
                y = [frac_vol]
            for j in range(len(org_sort_weights)):
                y.append(y[-1] - (org_sort_weights[j] / sum_weight)*frac_vol)
        else:
            y = np.ones(len(vox) + 1) - np.arange(0, len(vox) + 1) / len(vox)
        y[-1] = 0
        y = np.array(y)
        return x, y

    @staticmethod
    def get_max_dose(sol: dict, struct: str, dose_1d=None) -> float:
        """
        Get maximum dose_1d for the struct_name

        :param sol: optimal solution dictionary
        :param dose_1d: dose_1d which is not in solution dictionary
        :param struct: struct_name name

        :return: maximum dose_1d for the struct_name
        """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if len(vox) == 0:
            return 0
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        return np.max(dose_1d[vox])

    @staticmethod
    def get_mean_dose(sol: dict, struct: str, dose_1d=None) -> float:
        """
                Get mean dose_1d for the struct_name

                :param sol: optimal solution dictionary
                :param dose_1d: dose_1d which is not in solution dictionary
                :param struct: struct_name name

                :return: mean dose_1d for the struct_name
                """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if len(vox) == 0:
            return np.array(0)
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        frac_vol = inf_matrix.get_fraction_of_vol_in_calc_box(struct)
        mean_dose = (1 / sum(inf_matrix.get_opt_voxels_volume_cc(struct))) * (np.sum((np.multiply(inf_matrix.get_opt_voxels_volume_cc(struct), dose_1d[vox]))))
        return mean_dose * frac_vol

    @staticmethod
    def get_conformity_index(my_plan: Plan, sol: dict = None, dose_3d: np.ndarray = None, target_structure='PTV') -> float:
        """
        Calculate conformity index for the dose
        Closer to 1 is more better

        :param my_plan: object of class Plan
        :param sol: optimal solution dictionary
        :param dose_3d: dose in 3d array
        :param target_structure: target structure name

        :return: paddick conformity index

        """
        # calulating paddick conformity index
        percentile = 0.95  # reference isodose
        pres = my_plan.get_prescription()
        if dose_3d is None:
            dose_1d = sol['inf_matrix'].A @ (sol['optimal_intensity'] * my_plan.get_num_of_fractions())
            dose_3d = sol['inf_matrix'].dose_1d_to_3d(dose_1d=dose_1d)
        pres_iso_dose_mask = (dose_3d >= pres * percentile).astype(int)
        V_iso_pres = np.count_nonzero(pres_iso_dose_mask)
        ptv_mask = my_plan.structures.get_structure_mask_3d(target_structure)
        V_ptv = np.count_nonzero(ptv_mask)
        V_pres_iso_ptv = np.count_nonzero(pres_iso_dose_mask * ptv_mask)
        conformity_index = V_pres_iso_ptv * V_pres_iso_ptv / (V_ptv * V_iso_pres)
        return conformity_index

    @staticmethod
    def get_homogeneity_index(my_plan: Plan, sol: dict = None, dose_3d: np.ndarray = None, target_structure='PTV') -> float:
        """
                Calculate homogeneity index for the dose
                Closer to 0 is more better

                :param my_plan: object of class Plan
                :param sol: optimal solution dictionary
                :param dose_3d: dose in 3d array
                :param target_structure: target structure name

                :return: homogeneity index

                """
        if dose_3d is None:
            dose_1d = sol['inf_matrix'].A @ (sol['optimal_intensity'] * my_plan.get_num_of_fractions())
            dose_3d = sol['inf_matrix'].dose_1d_to_3d(dose_1d=dose_1d)
        ptv = my_plan.structures.get_structure_mask_3d(target_structure)
        ptv_dose = dose_3d[np.where(ptv == 1)]
        PTV_D2 = np.percentile(ptv_dose, 98)
        PTV_D50 = np.percentile(ptv_dose, 50)
        PTV_D98 = np.percentile(ptv_dose, 2)
        return (PTV_D2 - PTV_D98) / PTV_D50

    @staticmethod
    def get_BED(my_plan: Plan, sol: dict = None, dose_per_fraction_1d: np.ndarray = None, alpha=1, beta=1) -> np.ndarray:
        """
        Get Biologically equivalent dose (BED) for the struct_name

        :param my_plan: Object of class Plan
        :param sol: optimal solution dictionary
        :param dose_per_fraction_1d: dose_1d which is not in solution dictionary

        :return: BED
        """

        if dose_per_fraction_1d is None:
            dose_per_fraction_1d = sol['inf_matrix'].A @ (sol['optimal_intensity'])
        bed_d = np.zeros_like(dose_per_fraction_1d)
        for i in range(int(my_plan.get_num_of_fractions())):
            bed_d = bed_d + (dose_per_fraction_1d + (dose_per_fraction_1d**2/(alpha/beta)))
        return bed_d

    @staticmethod
    def add_dvh_to_frame(my_plan: Plan, df: pd.DataFrame, new_column_name: str, old_column_name: str, unit: str, dvh_type=None):
        req_ind = df.index[~df[old_column_name].isnull()].tolist()
        for ind in req_ind:
            if dvh_type is None:
                df.loc[ind, new_column_name] = str(round(Evaluation.get_num(my_plan, df[old_column_name][ind]), 2)) + unit
            elif dvh_type is not None:
                if dvh_type == 'dose_volume_V':
                    df.loc[ind, new_column_name] = 'V(' + str(round(Evaluation.get_num(my_plan, df[old_column_name][ind]), 2)) + unit + ')'
                elif dvh_type == 'dose_volume_D':
                    df.loc[ind, new_column_name] = 'D(' + str(round(Evaluation.get_num(my_plan, df[old_column_name][ind]), 2)) + unit + ')'
        return df

    @staticmethod
    def get_num(my_plan, string: Union[str, float]):
        if "prescription_gy" in str(string):
            prescription_gy = my_plan.get_prescription()
            return eval(string)
        elif isinstance(string, float) or isinstance(string, int):
            return string
        else:
            raise Exception('Invalid constraint')
    @staticmethod
    def is_notebook() -> bool:
        try:
            from IPython import get_ipython
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                if 'google.colab' in str(get_ipython()):
                    return True
                else:
                    return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter
        except:
            return False  # Probably standard Python interpreter



"""

This leaf sequencing algorithm is based upon intensity modulated beams with multiple static segments according to
Siochi (1999) International Journal of Radiation Oncology * Biology * Physics. This code has been inspired by
MatRad implementation by Eric Christiansen, Emily Heath, and Tong Xu.


"""
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from portpy.photon.plan import Plan
import numpy as np
from copy import deepcopy


def leaf_sequencing_siochi(my_plan: Plan, sol: dict, num_of_levels: int = 40) -> dict:
    """

    Create leaf sequences based upon optimal sol and returns dictionary containin leaf position and MU.

    :param my_plan: object of class Plan
    :param sol: solution dictionary
    :param num_of_levels: number of stratification levels. Default 40
    """

    beams = my_plan.beams
    inf_matrix = my_plan.inf_matrix
    leaf_sequencing = {}
    w_approx = np.zeros(inf_matrix.A.shape[1])
    for b, beam_id in enumerate(beams.get_all_beam_ids()):
        gantry_angle = beams.get_gantry_angle(beam_id=beam_id)
        leaf_sequencing[gantry_angle] = {}

        maps = inf_matrix.get_bev_2d_grid(beam_id=beam_id, finest_grid=False)
        numRows = np.size(maps, 0)
        numCols = np.size(maps, 1)
        w_maps = np.zeros((numRows, numCols))
        # wMaps = np.zeros((numRows, numCols))
        for r in range(numRows):
            for c in range(numCols):
                if maps[r, c] >= 0:
                    curr = maps[r, c]
                    w_maps[r, c] = sol['optimal_intensity'][curr]

        # create levels for fluence
        cal_fac = np.amax(w_maps)
        w_maps[w_maps == -1] = 0
        D_k = np.round(w_maps / cal_fac * num_of_levels)

        # Save

        D_0 = deepcopy(D_k)
        shapes = np.zeros((10000, D_k.shape[0], D_k.shape[1]), dtype='int')
        shapes_weight = np.zeros(10000)
        k = 0

        ind = np.where(D_k > 0)
        min_row = ind[0].min()
        max_row = ind[0].max()
        min_col = ind[1].min()
        max_col = ind[1].max()

        tops = np.zeros(D_k.shape, dtype='int')
        bases = np.zeros(D_k.shape, dtype='int')

        for i in range(min_col, max_col + 1):
            max_top = -1
            t_and_g = 1
            for j in range(min_row, max_row + 1):
                if i == min_col:
                    bases[j, i] = 1
                    tops[j, i] = bases[j, i] + D_k[j, i] - 1
                else:
                    if D_k[j, i] >= D_k[j, i - 1]:  # current rod greater than previous rod
                        bases[j, i] = bases[j, i - 1]
                        tops[j, i] = bases[j, i] + D_k[j, i] - 1
                    else:
                        if D_k[j, i] == 0:# rod length=0, put in in next slab after top of previous
                            bases[j, i] = tops[j, i - 1] + 1
                            tops[j, i] = bases[j, i] - 1
                        else:  # rod length ~ = 0, match tops
                            tops[j, i] = tops[j, i - 1]
                            bases[j, i] = tops[j, i] - D_k[j, i] + 1
                # determine which rod has the highest top in column
                if tops[j, i] > max_top:
                    max_top = tops[j, i]
                    top_row = j
            # check for collision and tongue and groove error
            while t_and_g:
                for j in range(top_row - 1, min_row-1, -1):
                    if D_k[j, i] < D_k[j + 1, i]:
                        if tops[j, i] > tops[j + 1, i]:
                            tops[j + 1, i] = tops[j, i]
                            bases[j + 1, i] = tops[j + 1, i] - D_k[j + 1, i] + 1
                        elif bases[j, i] < bases[j + 1, i]:
                            bases[j, i] = bases[j + 1, i]
                            tops[j, i] = bases[j, i] + D_k[j, i] - 1
                    else:
                        if tops[j, i] < tops[j + 1, i]:
                            tops[j, i] = tops[j + 1, i]
                            bases[j, i] = tops[j, i] - D_k[j, i] + 1
                        elif bases[j, i] > bases[j + 1, i]:
                            bases[j + 1, i] = bases[j, i]
                            tops[j + 1, i] = bases[j + 1, i] + D_k[j + 1, i] - 1

                # checking from max row to up
                for j in range(top_row + 1, max_row + 1):
                    if D_k[j, i] < D_k[j - 1, i]:
                        if tops[j, i] > tops[j - 1, i]:
                            tops[j - 1, i] = tops[j, i]
                            bases[j - 1, i] = tops[j - 1, i] - D_k[j - 1, i] + 1
                        elif bases[j, i] < bases[j - 1, i]:
                            bases[j, i] = bases[j - 1, i]
                            tops[j, i] = bases[j, i] + D_k[j, i] - 1
                    else:
                        if tops[j, i] < tops[j - 1, i]:
                            tops[j, i] = tops[j - 1, i]
                            bases[j, i] = tops[j, i] - D_k[j, i] + 1
                        elif bases[j, i] > bases[j - 1, i]:
                            bases[j - 1, i] = bases[j, i]
                            tops[j - 1, i] = bases[j - 1, i] + D_k[j - 1, i] - 1

                # check if t and g has been removed
                t_and_g = 0
                for j in range(min_row + 1, max_row + 1):
                    if D_k[j, i] < D_k[j - 1, i]:
                        if tops[j, i] > tops[j - 1, i]:
                            t_and_g = 1
                        elif bases[j, i] < bases[j-1, i]:
                            t_and_g = 1

                    else:
                        if tops[j, i] < tops[j - 1, i]:
                            t_and_g = 1
                        elif bases[j, i] > bases[j - 1, i]:
                            t_and_g = 1

        # convert to segments
        levels = np.amax(tops)
        for level in range(1, levels + 1):
            if siochi_diff_slab(tops, bases, level):
                shape_k = np.multiply(bases <= level, level <= tops)
                shapes[k, :, :] = shape_k
                k = k + 1
            shapes_weight[k-1] = shapes_weight[k-1] + 1

        leaf_shapes = shapes[0:k, :, :]
        leaf_weight = shapes_weight[0:k] / num_of_levels * cal_fac

        y_leaf_pos = beams.beams_dict['MLC_leaves_pos_y_mm'][0]
        sum_of_beam = np.zeros((numRows, numCols))
        for l in range(leaf_shapes.shape[0]):
            leaf_pos = np.empty((len(y_leaf_pos), 2), dtype='object')
            control_point = leaf_shapes[l, :, :]
            beamlet_idx_2d = control_point*(maps+1)
            beamlet_idx_2d = beamlet_idx_2d - 1  # adding and subtract 1
            for i in range(control_point.shape[0]):
                row = beamlet_idx_2d[i, :][beamlet_idx_2d[i, :] >= 0]

                if len(row) > 0:
                    first_beamlet = inf_matrix.get_beamlet_info(row[0])
                    last_beamlet = inf_matrix.get_beamlet_info(row[-1])
                    left_leaf_position = first_beamlet['position_x_mm'][0] - first_beamlet['width_mm'][0]/2
                    right_leaf_position = last_beamlet['position_x_mm'][0] + last_beamlet['width_mm'][0] / 2
                    ind = np.where(y_leaf_pos == int(first_beamlet['position_y_mm'][0]))[0][0]
                    ind_rev = len(y_leaf_pos) - ind - 1
                    leaf_pos[ind_rev, 0] = left_leaf_position
                    leaf_pos[ind_rev, 1] = right_leaf_position
            leaf_pos[:, 0][leaf_pos[:, 0] == None] = 43.5
            leaf_pos[:, 1][leaf_pos[:, 1] == None] = 44
            a = leaf_pos.flatten('F')
            string_list = [str(item) for item in a.tolist()]
            leaf_sequencing[gantry_angle].setdefault('leaf_postions', []).append(string_list)
            leaf_sequencing[gantry_angle].setdefault('MU', []).append(leaf_weight[l])
            leaf_sequencing[gantry_angle].setdefault('leaf_shapes', []).append(leaf_shapes[l])
            sum_of_beam = sum_of_beam + leaf_shapes[l]*leaf_weight[l]
        for r in range(numRows):
            for c in range(numCols):
                if maps[r, c] >= 0:
                    curr = maps[r, c]
                    w_approx[curr] = sum_of_beam[r, c]
    leaf_sequencing['optimal_intensity'] = w_approx
    return leaf_sequencing


def siochi_diff_slab(tops, bases, level):
    if level == 1:
        diff_slab = 1
    else:
        shape_level = np.multiply((bases <= level), level <= tops)
        shape_level_prev = np.multiply((bases <= level-1), (level-1 <= tops))
        diff_slab = not np.array_equiv(shape_level, shape_level_prev)
    return diff_slab




from .plan import Plan
from .structures import Structures
from .beam import Beams
from .ct import CT
from .influence_matrix import InfluenceMatrix
from .data_explorer import DataExplorer
from .optimization import Optimization
from .visualization import Visualization
from .evaluation import Evaluation
from .clinical_criteria import ClinicalCriteria
from portpy.photon.utils import *
try:
    from portpy.photon.vmat_scp import *
except ImportError:
    pass



# Read dicom files from dicom directory
from __future__ import annotations
import os
import SimpleITK as sitk
from pydicom import dcmread
import numpy as np
from portpy.photon.ct import CT
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from portpy.photon.plan import Plan


def read_dicom_dose(dir_name: str = None, dose_file_name: str = None):
    if dir_name is not None:
        dicom_names = os.listdir(dir_name)
        for dcm in dicom_names:
            if 'rt_dose' in dcm:
                dose_file_name = os.path.join(dir_name, dcm)

    dose_img = dcmread(dose_file_name)
    # dose_img = sitk.ReadImage([dcm for dcm in dicom_paths])
    arr_dose = dose_img.pixel_array
    rt_dose = arr_dose * dose_img.DoseGridScaling
    rt_dose_itk = sitk.GetImageFromArray(rt_dose)
    if 'Phantom' in dose_file_name:
        # There is issue for phantom patient. CT and dose dicom images have different origin w.r.t eclipse.
        # Eclipse have different origin
        dose_img.ImagePositionPatient[0] = dose_img.ImagePositionPatient[0] - 65.63
    rt_dose_itk.SetOrigin(dose_img.ImagePositionPatient)
    rt_dose_itk.SetSpacing([float(dose_img.PixelSpacing[0]), float(dose_img.PixelSpacing[1]),
                            dose_img.GridFrameOffsetVector[1] - dose_img.GridFrameOffsetVector[0]])
    # return dose_img
    return rt_dose_itk


def read_dicom(in_dir):
    dicom_names = os.listdir(in_dir)
    dicom_paths = []
    for dcm in dicom_names:
        if dcm[:2] == 'CT':
            dicom_paths.append(os.path.join(in_dir, dcm))

    img_positions = []
    for dcm in dicom_paths:
        ds = dcmread(dcm)
        img_positions.append(ds.ImagePositionPatient[2])

    indexes = np.argsort(np.asarray(img_positions))
    dicom_names = list(np.asarray(dicom_paths)[indexes])

    reader = sitk.ImageSeriesReader()
    reader.SetFileNames(dicom_names)
    img = reader.Execute()

    return img


def resample(img, ref_image):
    resampler = sitk.ResampleImageFilter()
    resampler.SetInterpolator(sitk.sitkLinear)
    resampler.SetReferenceImage(ref_image)
    img = resampler.Execute(img)

    return img


def get_ct_image(ct: CT):
    ct_arr = ct.ct_dict['ct_hu_3d'][0]
    ct_image = sitk.GetImageFromArray(ct_arr)
    ct_image.SetOrigin(ct.ct_dict['origin_xyz_mm'])
    ct_image.SetSpacing(ct.ct_dict['resolution_xyz_mm'])
    ct_image.SetDirection(ct.ct_dict['direction'])
    return ct_image


def convert_dose_rt_dicom_to_portpy(my_plan: Plan = None, ct: CT = None, dir_name: str = None, dose_file_name: str = None):
    dicom_dose_image = read_dicom_dose(dir_name=dir_name, dose_file_name=dose_file_name)
    if ct is None:
        ct = my_plan.ct
    ct_image = get_ct_image(ct)
    dose_3d_resampled_to_orgCT = resample(dicom_dose_image, ct_image)
    return sitk.GetArrayFromImage(dose_3d_resampled_to_orgCT)



from .base_model import BaseModel
from . import networks3d as networks
import torch


class TestModel(BaseModel):
    """ This TesteModel can be used to generate CycleGAN results for only one direction.
    This model will automatically set '--dataset_mode single', which only loads the images from one collection.

    See the test instruction for more details.
    """
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.

        The model can only be used during test time. It requires '--dataset_mode single'.
        You need to specify the network using the option '--model_suffix'.
        """
        assert not is_train, 'TestModel cannot be used during training time'
        parser.set_defaults(dataset_mode='single')
        parser.add_argument('--model_suffix', type=str, default='', help='In checkpoints_dir, [epoch]_net_G[model_suffix].pth will be loaded as the generator.')

        return parser

    def __init__(self, opt):
        """Initialize the pix2pix class.

        Parameters:
            opt (Option class)-- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        assert(not opt.isTrain)
        BaseModel.__init__(self, opt)
        # specify the training losses you want to print out. The training/test scripts  will call <BaseModel.get_current_losses>
        self.loss_names = ['G_MAE']
        # specify the images you want to save/display. The training/test scripts  will call <BaseModel.get_current_visuals>
        self.visual_names = ['real_CT', 'fake_Dose']
        # specify the models you want to save to the disk. The training/test scripts will call <BaseModel.save_networks> and <BaseModel.load_networks>
        self.model_names = ['G']  # only generator is needed.
        self.netG = networks.define_G(opt.input_nc, opt.output_nc, opt.ngf, opt.netG, opt.norm,
                                      not opt.no_dropout, opt.init_type, opt.init_gain, self.gpu_ids)

        # assigns the model to self.netG_[suffix] so that it can be loaded
        # please see <BaseModel.load_networks>
        # setattr(self, 'netG' + opt.model_suffix, self.netG)  # store netG in self.

    def set_input(self, input):
        """Unpack input data from the dataloader and perform necessary pre-processing steps.

        Parameters:
            input: a dictionary that contains the data itself and its metadata information.

        We need to use 'single_dataset' dataset mode. It only load images from one domain.
        """
        self.real_CT = input['A'].to(self.device)
        self.image_paths = input['A_paths']

    def forward(self):
        """Run forward pass."""
        self.fake_Dose = self.netG(self.real_CT)  # G(real)

    def optimize_parameters(self):
        """Calculate losses, gradients, and update network weights; called in every training iteration"""
        return 0

    def calculate_validation_loss(self):
        """Calculate losses, gradients, and update network weights; called in every training iteration"""
        return 0



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from portpy.photon.plan import Plan
    from portpy.photon.beam import Beams
from pydicom import dcmread
from pydicom import Dataset, Sequence
import numpy as np


def create_cp(my_plan, beam_id, cp_index, meterset_weight, leaf_positions):
    beams: Beams = my_plan.beams
    jaw_positions = beams.get_jaw_positions(beam_id)
    top_left_x_mm = jaw_positions['top_left_x_mm']
    bottom_right_x_mm = jaw_positions['bottom_right_x_mm']
    bottom_right_y_mm = jaw_positions['bottom_right_y_mm']
    top_left_y_mm = jaw_positions['top_left_y_mm']
    energy_mv = beams.beams_dict['energy_MV'][0]
    iso_center = beams.get_iso_center(beam_id)


    control_point = Dataset()
    control_point.ControlPointIndex = str(cp_index)
    control_point.NominalBeamEnergy = energy_mv
    control_point.DoseRateSet = 600  # hard coded
    BeamLimitingDevicePositionSequence = Sequence()

    beam_limiting_dataset_x = Dataset()
    # X jaw.[X1 X2] -ve of the X1 we see in eclipse
    beam_limiting_dataset_x.RTBeamLimitingDeviceType = 'ASYMX'
    beam_limiting_dataset_x.LeafJawPositions = [top_left_x_mm, bottom_right_x_mm+my_plan.beams.get_beamlet_width()] # Temporary bug fix
    BeamLimitingDevicePositionSequence.append(beam_limiting_dataset_x)

    # Y jaw.[Y1 Y2] -ve of the Y1 we see in eclipse
    beam_limiting_dataset_y = Dataset()
    beam_limiting_dataset_y.RTBeamLimitingDeviceType = 'ASYMY'
    beam_limiting_dataset_y.LeafJawPositions = [bottom_right_y_mm - my_plan.beams.get_beamlet_height(), top_left_y_mm]
    BeamLimitingDevicePositionSequence.append(beam_limiting_dataset_y)

    beam_limiting_dataset_leaf_positions = Dataset()
    beam_limiting_dataset_leaf_positions.RTBeamLimitingDeviceType = 'MLCX'
    beam_limiting_dataset_leaf_positions.LeafJawPositions = leaf_positions
    BeamLimitingDevicePositionSequence.append(beam_limiting_dataset_leaf_positions)

    control_point.BeamLimitingDevicePositionSequence = BeamLimitingDevicePositionSequence
    control_point.GantryAngle = beams.get_gantry_angle(beam_id)
    control_point.GantryRotationDirection = "CW"
    control_point.CumulativeMetersetWeight = meterset_weight

    # Additional attributes
    if cp_index == 0:
        control_point.BeamLimitingDeviceAngle = 0
        control_point.BeamLimitingDeviceRotationDirection = 'NONE'
        control_point.PatientSupportAngle = 0
        control_point.PatientSupportRotationDirection = 'NONE'
        control_point.TableTopEccentricAngle = 0.0
        control_point.TableTopEccentricRotationDirection = "NONE"
        control_point.TableTopVerticalPosition = 0.0
        control_point.TableTopLongitudinalPosition = 1000.0
        control_point.TableTopLateralPosition = 0.0
        control_point.IsocenterPosition = [iso_center['x_mm'], iso_center['y_mm'], iso_center['z_mm']]
        control_point.TableTopPitchAngle = 0.0
        control_point.TableTopPitchRotationDirection = "NONE"
        control_point.TableTopRollAngle = 0.0
        control_point.TableTopRollRotationDirection = "NONE"
    return control_point

def create_beam(my_plan, arc_id, beam_number):
    ind = [i for i in range(len(my_plan.arcs.arcs_dict['arcs'])) if
           my_plan.arcs.arcs_dict['arcs'][i]['arc_id'] == arc_id][0]
    leaf_y = [-200, -190, -180, -170, -160, -150, -140, -130, -120, -110, -100, -95, -90, -85, -80, -75, -70, -65, -60, -55, -50, -45, -40, -35, -30, -25, -20, -15, -10, -5, 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110, 120, 130, 140, 150, 160, 170, 180, 190, 200]
    beam = Dataset()
    beam.Manufacturer = "Varian Medical Systems"
    beam.ManufacturerModelName = "TDS"
    beam.DeviceSerialNumber = "1003"

    # Primary Fluence Mode Sequence
    primary_fluence_mode = Dataset()
    primary_fluence_mode.FluenceMode = "STANDARD"
    beam.PrimaryFluenceModeSequence = [primary_fluence_mode]

    beam.TreatmentMachineName = "445HET1"
    beam.PrimaryDosimeterUnit = "MU"
    beam.SourceAxisDistance = my_plan.beams.beams_dict['SAD_mm'][0]

    # Beam Limiting Device Sequence
    beam_limiting_device_sequence = Sequence()

    # ASYMX
    asxmx = Dataset()
    asxmx.RTBeamLimitingDeviceType = "ASYMX"
    asxmx.NumberOfLeafJawPairs = "1"
    beam_limiting_device_sequence.append(asxmx)

    # ASYMY
    asxmy = Dataset()
    asxmy.RTBeamLimitingDeviceType = "ASYMY"
    asxmy.NumberOfLeafJawPairs = "1"
    beam_limiting_device_sequence.append(asxmy)

    # MLCX
    mlcx = Dataset()
    mlcx.RTBeamLimitingDeviceType = "MLCX"
    mlcx.SourceToBeamLimitingDeviceDistance = "509.0"  # Hard coded. modify it
    mlcx.NumberOfLeafJawPairs = "60"
    mlcx.LeafPositionBoundaries = leaf_y
    beam_limiting_device_sequence.append(mlcx)

    beam.BeamLimitingDeviceSequence = beam_limiting_device_sequence

    beam.BeamNumber = str(beam_number)
    beam.BeamName = f"Field {arc_id}"
    beam.BeamType = "DYNAMIC"
    beam.RadiationType = "PHOTON"
    beam.TreatmentDeliveryType = "TREATMENT"
    beam.NumberOfWedges = "0"
    beam.NumberOfCompensators = "0"
    beam.NumberOfBoli = "0"
    beam.NumberOfBlocks = "0"
    beam.FinalCumulativeMetersetWeight = np.round(1.0, 4)
    beam.NumberOfControlPoints = str(len(my_plan.arcs.arcs_dict['arcs'][ind]['vmat_opt']))
    beam.ReferencedPatientSetupNumber = beam_number
    return beam

def edit_fraction_group_sequence(my_plan, ds):
    ds.FractionGroupSequence[0].NumberOfBeams = str(len(my_plan.arcs.arcs_dict['arcs']))
    from pydicom.uid import generate_uid

    reference_dose_uid = generate_uid()
    # Define Referenced Beam Sequence
    referenced_beam_sequence = Sequence()
    patient_setup_sequence = Sequence()
    for i in range(len(my_plan.arcs.arcs_dict['arcs'])):
        # Add Referenced Beam attributes
        referenced_beam = Dataset()
        referenced_beam.ReferencedDoseReferenceUID = reference_dose_uid
        referenced_beam.ReferencedBeamNumber = str(i+1)

        # Add Referenced Beam to sequence
        referenced_beam_sequence.append(referenced_beam)
        # Create Patient Setup attributes
        patient_setup = Dataset()
        patient_setup.PatientPosition = "HFS"
        patient_setup.PatientSetupNumber = str(i+1)
        patient_setup.SetupTechnique = "ISOCENTRIC"

        # Add Patient Setup to sequence
        patient_setup_sequence.append(patient_setup)

    # Add Patient Setup Sequence to RT Plan
    ds.PatientSetupSequence = patient_setup_sequence
    ds.FractionGroupSequence[0].ReferencedBeamSequence = referenced_beam_sequence

    # refer new dose
    dose_reference_sequence = Sequence()
    # Create Dose Reference attributes
    dose_reference = Dataset()
    dose_reference.DoseReferenceNumber = "1"
    dose_reference.DoseReferenceUID = reference_dose_uid
    # Add Dose Reference to sequence
    dose_reference_sequence.append(dose_reference)
    # Add Dose Reference Sequence to RT Plan
    ds.DoseReferenceSequence = dose_reference_sequence

    return ds
def write_rt_plan_vmat(my_plan: Plan, out_rt_plan_file: str, in_rt_plan_file: str):
    """

    Write output from leaf sequencing to dicom rt plan

    :param my_plan: object of class plan
    :param in_rt_plan_file: file location of input rt plan containing imrt beams
    :param out_rt_plan_file: new rt plan which can be imported in TPS

    """

    # read rt plan file using pydicom
    ds = dcmread(in_rt_plan_file)
    del ds.BeamSequence
    arcs = my_plan.arcs
    for a, arc in enumerate(arcs.arcs_dict['arcs']):
        for b, beam in enumerate(arc['vmat_opt']):
            beam['field_size'] = beam['best_leaf_position_in_cm'][:, 1]*10 - beam['best_leaf_position_in_cm'][:, 0]*10
            beam_id = beam['beam_id']
            beamlets = my_plan.beams.get_beamlets(beam_id=beam_id)
            min_origin_x = np.min(beamlets['position_x_mm'])
            beam['bank_b'] = min_origin_x + beam['best_leaf_position_in_cm'][:, 0]*10 - my_plan.beams.get_finest_beamlet_width() / 2
            beam['bank_a'] = min_origin_x + beam['best_leaf_position_in_cm'][:, 1]*10 + my_plan.beams.get_finest_beamlet_width() / 2

    # Write bank positions to text file
    # the leaf pairs between max_leaf_pair and min_leaf_pair must satisfy the dynamic min leaf gap constraint
    # min gap between leaf pairs should be at least 0.5 mm
    total_lp = 60  # Temporary. Tell Hai to add it to metadata
    ds.BeamSequence = Sequence()
    # ds = edit_fraction_group_sequence(my_plan, ds)
    for a, arc in enumerate(arcs.arcs_dict['arcs']):
        arc_id = arc['arc_id']
        beam_dataset = create_beam(my_plan=my_plan, arc_id=arc_id, beam_number=a+1)
        mu = [arc['vmat_opt'][i]['best_beam_weight'] for i in range(len(arc['vmat_opt']))]
        mu[0] = 0  # make first beam 0 mu
        total_mu = sum(mu)
        meterset_weight = 0
        control_point_sequence = Sequence()
        for b, beam in enumerate(arc['vmat_opt']):
            beam_id = beam['beam_id']
            leaf_pos = np.empty((total_lp, 2), dtype='object')
            stop = int(beam['start_leaf_pair']) - 1  # We are doing it in reverse way since rt plan have reverse index list
            for r in range(beam['num_rows']):
                start = stop - len(beam['MLC_leaf_idx'][0][0])
                if beam['field_size'][r] < 0.5:
                    leaf_pos[start:stop, 1] = [0]
                else:
                    leaf_pos[start:stop, 0] = [beam['bank_b'][r]]
                    leaf_pos[start:stop, 1] = [beam['bank_a'][r]]
                stop = stop - len(beam['MLC_leaf_idx'][0][0])

            leaf_pos[:, 0][leaf_pos[:, 0] == None] = 0.5
            leaf_pos[:, 1][leaf_pos[:, 1] == None] = 0.5
            a = leaf_pos.flatten('F')
            string_list = [str(item) for item in a.tolist()]
            if b > 0:
                meterset_weight += beam['best_beam_weight'] / total_mu
            control_point = create_cp(my_plan, beam_id, b, meterset_weight, string_list)
            control_point_sequence.append(control_point)
        beam_dataset.ControlPointSequence = control_point_sequence
        ds.BeamSequence.append(beam_dataset)
    ds.save_as(out_rt_plan_file)
    print('Dicom rt plan file created..')



from data.base_dataset import BaseDataset, get_transform
from data.image_folder import make_dataset
import torch
import numpy as np
from .base_dataset import transform_3d_data_test
import os

class SingleDataset(BaseDataset):
    """This dataset class can load a set of images specified by the path --dataroot /path/to/data.

    It can be used for generating CycleGAN results only for one side with the model option '-model test'.
    """

    def __init__(self, opt):
        """Initialize this dataset class.

        Parameters:
            opt (Option class) -- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseDataset.__init__(self, opt)
        self.dir_A = os.path.join(opt.dataroot)  # get the directory where images are located
        # self.dir_A = os.path.join(opt.dataroot, opt.phase)  # get the image directory
        self.A_paths = sorted(make_dataset(self.dir_A, opt.max_dataset_size))  # get image paths
        # self.A_paths = sorted(make_dataset(opt.dataroot, opt.max_dataset_size))
        self.input_nc = self.opt.output_nc if self.opt.direction == 'BtoA' else self.opt.input_nc
        # self.transform = get_transform(opt, grayscale=(input_nc == 1))

    def __getitem__(self, index):
        """Return a data point and its metadata information.

        Parameters:
            index - - a random integer for data indexing

        Returns a dictionary that contains A and A_paths
            A(tensor) - - an image in one domain
            A_paths(str) - - the path of the image
        """
        A_path = self.A_paths[index]
        AB = np.load(A_path)  ##Added (Gourav)

        # Get AB image into A and B
        A = AB['CT']
        OAR = AB['OAR']
        PTV = AB['PTV']
        # hist = AB['HIST']
        # bins = AB['BINS']
        beam = AB['BEAM']

        A, OAR, beam = transform_3d_data_test(A, OAR, PTV, beam, transform=False)
        A = torch.unsqueeze(A, dim=0)  # Add channel dimensions as data is 3D
        # B = torch.unsqueeze(B, dim=0)
        beam = torch.unsqueeze(beam, dim=0)

        A = torch.cat((A, beam, OAR), axis=0)
        return {'A': A, 'A_paths': A_path}

    def __len__(self):
        """Return the total number of images in the dataset."""
        return len(self.A_paths)



import numpy as np
from typing import List, Union
from .data_explorer import DataExplorer


class Beams:
    """
    A class representing beams_dict.

    - **Attributes** ::

        :param beams_dict: beams_dict dictionary that contains information about the beams_dict in the format of
        dict: {
                   'ID': list(int),
                   'gantry_angle': list(float),
                   'collimator_angle': list(float) }
                  }

        :type beams_dict: dict

    - **Methods** ::

        :get_gantry_angle(beam_id: Optional(int, List[int]):
            Get gantry angle in degrees
        :get_collimator_angle(beam_id):
            Get collimator angle in degrees

    """

    def __init__(self, data: DataExplorer, beam_ids:  List[Union[int, str]] = None, load_inf_matrix_full: bool = False):
        """

        :param beams_dict: Beams dictionary containing information about beams
        """

        metadata = data.load_metadata()
        metadata = self.get_plan_beams(beam_ids=beam_ids, meta_data=metadata)
        beams_dict = data.load_data(meta_data=metadata['beams'], load_inf_matrix_full=load_inf_matrix_full)
        self.beams_dict = beams_dict
        self.preprocess_beams()

    def get_beamlet_idx_2d_finest_grid(self, beam_id: Union[int, str]) -> np.ndarray:
        """

        :param beam_id: beam_id for the beam
        :return: 2d grid of beamlets in 2.5*2.5 resolution
        """
        ind = self.beams_dict['ID'].index(beam_id)
        return self.beams_dict['beamlet_idx_2d_finest_grid'][ind]

    def get_gantry_angle(self, beam_id: Union[Union[int, str], List[Union[int, str]]]) -> Union[int, List[int]]:
        """
        Get gantry angle

        :param beam_id: beam_id for the beam
        :return: gantry angle for the beam_id
        """
        if isinstance(beam_id, int) or isinstance(beam_id, str):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['gantry_angle'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['gantry_angle'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    def get_collimator_angle(self, beam_id: Union[Union[int, str], List[Union[int, str]]]) -> Union[int, List[int]]:
        """
        Get collimator angle

        :param beam_id: beam_id for the beam
        :return: collimator angle for the beam_id
        """
        if isinstance(beam_id, int) or isinstance(beam_id, str):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['collimator_angle'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['collimator_angle'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    def get_jaw_positions(self, beam_id: Union[Union[int, str], List[Union[int, str]]]) -> Union[dict, List[dict]]:
        """
        Get jaw positions

        :param beam_id: beam_id for the beam
        :return: jaw position for the beam_id
        """
        if isinstance(beam_id, int) or isinstance(beam_id, str):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['jaw_position'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['jaw_position'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    def get_iso_center(self, beam_id: Union[Union[int, str], List[Union[int, str]]]) -> Union[dict, List[dict]]:
        """
        Get iso center for the given beam_id/ids

        :param beam_id: beam_id for the beam
        :return: iso center for the beam_id
        """
        if isinstance(beam_id, int) or  isinstance(beam_id, str):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['iso_center'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['iso_center'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    def get_beamlets(self, beam_id: Union[Union[int, str], List[Union[int, str]]]) -> Union[dict, List[dict]]:
        """
        Get jaw positions

        :param beam_id: beam_id for the beam
        :return: jaw position for the beam_id
        """
        if isinstance(beam_id, int) or isinstance(beam_id, str):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['beamlets'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['beamlets'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    @staticmethod
    def sort_beamlets(b_map):
        c = b_map[b_map >= 0]
        c = np.unique(c)
        ind = np.arange(0, len(c))
        c_sort = np.sort(c)
        matrix_ind = [np.where(b_map == c_i) for c_i in c_sort]
        map_copy = b_map.copy()
        for i in range(len(ind)):
            map_copy[matrix_ind[i]] = ind[i]
        return map_copy

    def make_beamlets_continous(self):

        # make beamlets continuous

        for ind, id in enumerate(self.beams_dict['ID']):
            beam_map = self.get_beamlet_idx_2d_finest_grid(beam_id=id)
            std_map = self.sort_beamlets(beam_map)
            if ind == 0:
                self.beams_dict['beamlet_idx_2d_finest_grid'][ind] = std_map
            else:

                beamlet_ind = np.unique(std_map.flatten())
                beamlet_ind = beamlet_ind[beamlet_ind >= 0]
                a = np.where(np.isin(std_map, beamlet_ind))
                mask_2d_grid = np.zeros_like(std_map, dtype=bool)
                mask_2d_grid[a] = True
                beam_map = std_map + int(
                    np.amax(self.beams_dict['beamlet_idx_2d_finest_grid'][ind - 1]) + 1) * mask_2d_grid
                self.beams_dict['beamlet_idx_2d_finest_grid'][ind] = beam_map

    def preprocess_beams(self):
        for i, beam_id in enumerate(self.beams_dict['ID']):
            ind = self.beams_dict['ID'].index(beam_id)
            beam_2d_grid = self.create_beamlet_idx_2d_finest_grid(beam_id=beam_id)
            self.beams_dict.setdefault('beamlet_idx_2d_finest_grid', []).append(beam_2d_grid)

    @staticmethod
    def _create_2d_orig_grid(beam_map):
        rowsNoRepeat = [0]
        for i in range(1, np.size(beam_map, 0)):
            if (beam_map[i, :] != beam_map[rowsNoRepeat[-1], :]).any():
                rowsNoRepeat.append(i)
        colsNoRepeat = [0]
        for j in range(1, np.size(beam_map, 1)):
            if (beam_map[:, j] != beam_map[:, colsNoRepeat[-1]]).any():
                colsNoRepeat.append(j)
        beam_map = beam_map[np.ix_(np.asarray(rowsNoRepeat), np.asarray(colsNoRepeat))]
        return beam_map

    def get_beamlet_idx_2d_grid(self, beam_id: Union[Union[int, str], List[Union[int, str]]] = None, ind: int = None,
                                finest_grid: bool = False) -> Union[np.ndarray, List[np.ndarray]]:

        """
        get BEV in 2d grid in original resolution where beamlet index is the column number in influence matrix

        :param ind: idx of the beam in beamlets_dict.
        It can be int or List[int]. If int, ndarray is return, If List[int], list[ndarray] is returned
        :param beam_id: beam_id of the beam.
        It can be int or List[int]. If int, ndarray is return, If List[int], list[ndarray] is returned
        :param finest_grid:Default to False. If set to true, it will return 2d grid in finest resolution
        :return: ndarray/List[ndarray]
        """

        if beam_id is not None:
            ind = []
            if isinstance(beam_id, int) or isinstance(beam_id, str):
                ind = [i for i in range(len(self.beams_dict['ID'])) if
                       self.beams_dict['ID'][i] == beam_id]
                if not ind:
                    raise ValueError("invalid beam id {}".format(beam_id))
            elif isinstance(beam_id, list):
                for idx in beam_id:
                    try:
                        ind_1 = [i for i in range(len(self.beams_dict['ID'])) if
                                 self.beams_dict['ID'][i] == idx][0]
                        ind.append(ind_1)
                    except:
                        raise ValueError("invalid beam id {}".format(idx))
        # Bug fix. in case ind is int make it as list
        if isinstance(ind, int) or isinstance(ind, str):
            ind = [ind]
        beam_orig = []
        for b in ind:
            beam_map = self.beams_dict['beamlet_idx_2d_finest_grid'][b]
            if not finest_grid:
                beam_map = self._create_2d_orig_grid(beam_map)
            beam_orig.append(beam_map)
        if len(beam_orig) == 1:
            beam_orig = beam_orig[0]  # return ndarray in case if it is not list

        return beam_orig

    @staticmethod
    def get_finest_beamlet_width() -> float:
        """

        :return: beamlet width in the original beam
        """

        return 2.5

    @staticmethod
    def get_finest_beamlet_height() -> float:
        """

        :return: beamlet height in the original beam
        """

        return 2.5

    def get_beamlet_width(self) -> float:
        """

        :return: beamlet width in the original beam
        """
        beamlets = self.beams_dict['beamlets'][0]
        return np.squeeze(beamlets['width_mm'][0])[0]

    def get_beamlet_height(self) -> float:
        """

        :return: beamlet height in the original beam
        """
        beamlets = self.beams_dict['beamlets'][0]
        return np.squeeze(beamlets['height_mm'][0])[0]

    def get_all_beam_ids(self) -> List[int]:
        return self.beams_dict['ID']

    def create_beamlet_idx_2d_finest_grid(self, beam_id: int) -> np.ndarray:
        """
        Create 2d grid for the beamlets where each element is 2.5mm*2.5mm for the given beam id from x and y coordinates of beamlets.

        :param beam_id: beam_id for the beam
        :return: 2d grid of beamlets for the beam

        """
        ind = self.beams_dict['ID'].index(beam_id)
        beamlets = self.beams_dict['beamlets'][ind]
        x_positions = beamlets['position_x_mm'][0] - beamlets['width_mm'][0]/2  # x position is center of beamlet. Get left corner
        y_positions = beamlets['position_y_mm'][0] + beamlets['height_mm'][0]/2  # y position is center of beamlet. Get top corner
        right_ind = np.argmax(x_positions)
        bottom_ind = np.argmin(y_positions)
        w_all = np.column_stack((x_positions, y_positions))  # top left corners of all beamlets
        x_coord = np.arange(np.min(x_positions), np.max(x_positions) + beamlets['width_mm'][0][right_ind].item(), 2.5)
        y_coord = np.arange(np.max(y_positions), np.min(y_positions) - beamlets['height_mm'][0][bottom_ind].item(), -2.5)

        # create mesh grid in 2.5 mm resolution
        XX, YY = np.meshgrid(x_coord, y_coord)
        beamlet_idx_2d_finest_grid = np.ones_like(XX, dtype=int)
        beamlet_idx_2d_finest_grid = beamlet_idx_2d_finest_grid * int(-1)  # make all elements to -1

        for row in range(XX.shape[0]):
            for col in range(XX.shape[1]):
                ind = np.where((w_all[:, 0] == XX[row, col]) & (w_all[:, 1] == YY[row, col]))  # find the position in matrix where we find the beamlet
                if np.size(ind) > 0:
                    ind = ind[0][0]
                    num_width = int(beamlets['width_mm'][0][ind].item() /2.5)
                    num_height = int(beamlets['height_mm'][0][ind].item() / 2.5)
                    beamlet_idx_2d_finest_grid[row:row+num_height, col:col+num_width] = ind
        return beamlet_idx_2d_finest_grid

    @staticmethod
    def get_plan_beams(beam_ids: List[Union[int, str]] = None, meta_data: dict = None) -> dict:
        """
        Create and return a copy of meta_data with only including the requested beams_dict (beam_ids)


        :param beam_ids: the indices of the beams_dict to be included. If None, the planner's beams_dict are used
        :param meta_data: the dictionary including all the beams_dict
        :return: returns the meta_data dictionary only including the requested beams_dict in format of:
            dict: {
                   'structures': {'name': list(str), 'volume_cc': list(float), }
                   'opt_voxels': {'_ct_voxel_resolution_xyz_mm': list(float),}
                  }
        """
        if beam_ids is None:  # if beam_ids not included, then the beams_dict
            # selected by an expert human planner would be used
            if 'planner_beam_ids' in meta_data:
                beam_ids = meta_data['planner_beam_ids']['IDs']
            else:
                beam_ids = meta_data['beams']['ID']
        meta_data_req = meta_data.copy()
        del meta_data_req['beams']  # remove previous beams_dict
        beamReq = dict()
        for i in range(len(beam_ids)):
            if beam_ids[i] in meta_data['beams']['ID']:
                ind = meta_data['beams']['ID'].index(beam_ids[i])
                for key in meta_data['beams']:
                    beamReq.setdefault(key, []).append(meta_data['beams'][key][ind])
            else:
                print('beam id {} is not available'.format(beam_ids[i]))
        meta_data_req['beams'] = beamReq
        return meta_data_req



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from portpy.photon.plan import Plan
from pydicom import dcmread
from pydicom import Dataset, Sequence


def write_rt_plan_imrt(my_plan: Plan, leaf_sequencing: dict, out_rt_plan_file: str, in_rt_plan_file: str):
    """

    Write output from leaf sequencing to dicom rt plan

    :param my_plan: object of class plan
    :param leaf_sequencing: dictionary containing leaf positions and shapes
    :param in_rt_plan_file: file location of input rt plan containing imrt beams
    :param out_rt_plan_file: new rt plan which can be imported in TPS

    """
    # get positions. PortPy have same jaw settings for all beams
    top_left_x_mm = my_plan.beams.beams_dict['jaw_position'][0]['top_left_x_mm']
    bottom_right_x_mm = my_plan.beams.beams_dict['jaw_position'][0]['bottom_right_x_mm']
    bottom_right_y_mm = my_plan.beams.beams_dict['jaw_position'][0]['bottom_right_y_mm']
    top_left_y_mm = my_plan.beams.beams_dict['jaw_position'][0]['top_left_y_mm']

    # read rt plan file using pydicom
    ds = dcmread(in_rt_plan_file)
    for b in range(len(ds.BeamSequence)):
        gantry_angle = int(ds.BeamSequence[b].ControlPointSequence[0].GantryAngle)
        meterset_weight = 0
        if gantry_angle in leaf_sequencing:
            leaf_postions = leaf_sequencing[gantry_angle]['leaf_postions']
            del ds.BeamSequence[b].ControlPointSequence[1:]  # delete all the control point after 1st control point
            ds.BeamSequence[b].NumberOfControlPoints = len(leaf_postions) + 1
            for cp, shape in enumerate(leaf_postions):
                ds.BeamSequence[b].ControlPointSequence.append(Dataset())
                ds.BeamSequence[b].ControlPointSequence[cp+1].ControlPointIndex = str(cp + 1)
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence = Sequence()
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence.append(Dataset())

                # X jaw.[X1 X2] -ve of the X1 we see in eclipse
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence[0].RTBeamLimitingDeviceType = 'ASYMX'
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence[0].LeafJawPositions = [top_left_x_mm, bottom_right_x_mm]

                # Y jaw.[Y1 Y2] -ve of the Y1 we see in eclipse
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence.append(Dataset())
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence[1].RTBeamLimitingDeviceType = 'ASYMY'
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence[1].LeafJawPositions = [bottom_right_y_mm, top_left_y_mm]

                # set leafs
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence.append(Dataset())
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence[2].RTBeamLimitingDeviceType = 'MLCX'
                ds.BeamSequence[b].ControlPointSequence[cp+1].BeamLimitingDevicePositionSequence[2].LeafJawPositions = shape

                meterset_weight = meterset_weight + leaf_sequencing[gantry_angle]['MU'][cp] / sum(leaf_sequencing[gantry_angle]['MU'])
                ds.BeamSequence[b].ControlPointSequence[cp+1].CumulativeMetersetWeight = round(meterset_weight, 4)

    ds.save_as(out_rt_plan_file)
